use std::path::PathBuf;

use serde::{
    Deserialize,
    Serialize,
    Serializer,
    ser::Error,
    ser::SerializeStruct,
};
use ssh_key::{
    Algorithm,
    LineEnding,
    private::PrivateKey,
    public::PublicKey,
};

use crate::errors::*;


#[derive(Debug, Clone, Eq, PartialEq)]
pub struct SshKey {
    pub path: String,
    pub algorithm: Algorithm,
    pub comment: String,
    pub public_key: PublicKey,
    pub private_key: PrivateKey,
}

impl SshKey {
    pub fn from_file(path: String, passphrase: &str) -> Result<SshKey, LoadSshKeyError> {
        let mut privkey = PrivateKey::read_openssh_file(path.as_ref())?;
        if privkey.is_encrypted() {
            privkey = privkey.decrypt(passphrase)
                .map_err(|_| LoadSshKeyError::InvalidPassphrase)?;
        }

        Ok(SshKey {
            path,
            algorithm: privkey.algorithm(),
            comment: privkey.comment().into(),
            public_key: privkey.public_key().clone(),
            private_key: privkey,
        })
    }
}


impl Serialize for SshKey {
    fn serialize<S: Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
        let mut key = s.serialize_struct("SshKey", 5)?;
        key.serialize_field("path", &self.path)?;
        key.serialize_field("algorithm", self.algorithm.as_str())?;
        key.serialize_field("comment", &self.comment)?;

        let pubkey_str = self.public_key.to_openssh()
            .map_err(|e| S::Error::custom(format!("Failed to encode SSH public key: {e}")))?;
        key.serialize_field("public_key", &pubkey_str)?;

        let privkey_str = self.private_key.to_openssh(LineEnding::LF)
            .map_err(|e| S::Error::custom(format!("Failed to encode SSH private key: {e}")))?;
        key.serialize_field::<str>("private_key", privkey_str.as_ref())?;

        key.end()
    }
}


#[cfg(test)]
mod tests {
    use super::*;
    use std::fs;

    fn path(name: &str) -> String {
        format!("./src/credentials/fixtures/{name}")
    }


    #[test]
    fn test_load_rsa_plain() {
        let k = SshKey::from_file(path("ssh_rsa_plain"), "")
            .expect("Failed to load SSH key");
    }


    #[test]
    fn test_load_rsa_enc() {
        let k = SshKey::from_file(
            path("ssh_rsa_enc"),
            "correct horse battery staple",
        ).expect("Failed to load SSH key");
    }


    #[test]
    fn test_load_ed25519_plain() {
        let k = SshKey::from_file(path("ssh_ed25519_plain"), "")
            .expect("Failed to load SSH key");
    }


    #[test]
    fn test_load_ed25519_enc() {
        let k = SshKey::from_file(
            path("ssh_ed25519_enc"),
            "correct horse battery staple",
        ).expect("Failed to load SSH key");
    }


    #[test]
    fn test_serialize() {
        let expected = fs::read_to_string(path("ssh_ed25519_plain.json")).unwrap();

        let k = SshKey::from_file(path("ssh_ed25519_plain"), "").unwrap();
        let computed = serde_json::to_string(&k)
            .expect("Failed to serialize SshKey");

        assert_eq!(expected, computed);
    }
}
